@@ -54,6 +54,8 @@
 #include "progname.h"
 #include "xalloc.h"
 
+#include "bind-mount.h"
+
 #ifdef __GNUC__
 #define attribute__(X) __attribute__ (X)
 #else
@@ -68,7 +70,9 @@
 #define LOG_FATAL	6
 
 #define FLAG_FORCE	(1 << 0)
-#define FLAG_ALLFILES	(1 << 1)   /* normally just files, dirs are removed */
+/* normally just files, dirs, and sockets older than (boot time - grace period)
+   are removed */
+#define FLAG_ALLFILES	(1 << 1)
 #define FLAG_TEST	(1 << 2)
 #define FLAG_ATIME	(1 << 3)
 #define FLAG_MTIME	(1 << 4)
@@ -108,6 +112,11 @@ struct excluded_uid
 static struct excluded_uid *excluded_uids /* = NULL */;
 static struct excluded_uid **excluded_uids_tail = &excluded_uids;
 
+static time_t kill_time;
+static time_t socket_kill_time; /* 0 = never */
+
+static int config_flags; /* = 0; */
+
 static int logLevel = LOG_NORMAL;
 
 static void attribute__((format(printf, 2, 3)))
@@ -117,14 +126,14 @@ static void attribute__((format(printf, 2, 3)))
     FILE * where = stdout;
 
     if (level >= logLevel) {
-	va_start(args, format);
-
 	if (level > LOG_NORMAL) {
 	    where = stderr;
 	    fprintf(where, "error: ");
 	}
 
+	va_start(args, format);
 	vfprintf(where, format, args);
+	va_end(args);
 
 	if (level == LOG_FATAL) exit(1);
     }
@@ -179,7 +188,7 @@ safe_chdir(const char *fulldirname, const char *reldirname, dev_t st_dev,
 	return 1;
     }
 
-    if (chdir(reldirname)) {
+    if (chdir(reldirname) != 0) {
 	if (errno == ENOENT)
 	    return 2;
 	message(LOG_ERROR, "chdir to directory %s failed: %s\n",
@@ -187,7 +196,7 @@ safe_chdir(const char *fulldirname, const char *reldirname, dev_t st_dev,
 	return 1;
     }
 
-    if (lstat(".", &sb2)) {
+    if (lstat(".", &sb2) != 0) {
 	message(LOG_ERROR, "second lstat() of directory %s failed: %s\n",
 		fulldirname, strerror(errno));
 	return 1;
@@ -242,7 +251,7 @@ check_fuser(const char *filename)
 	waitpid(pid, &ret, 0);
     }
 
-    return (WIFEXITED(ret) && (WEXITSTATUS(ret) == 0));
+    return (WIFEXITED(ret) && WEXITSTATUS(ret) == 0);
 }
 #else
 #define check_fuser(FILENAME) 0
@@ -284,8 +293,8 @@ is_mount_point(const char *path)
 #endif
 
 static int
-cleanupDirectory(const char * fulldirname, const char *reldirname,
-		 time_t killTime, int flags, dev_t st_dev, ino_t st_ino)
+cleanupDirectory(const char * fulldirname, const char *reldirname, dev_t st_dev,
+		 ino_t st_ino)
 {
     DIR *dir;
     struct dirent *ent;
@@ -308,7 +317,7 @@ cleanupDirectory(const char * fulldirname, const char *reldirname,
 	return 1;
     }
 
-    if (lstat(".", &here)) {
+    if (lstat(".", &here) != 0) {
 	message(LOG_ERROR, "error stat()ing current directory %s: %s\n",
 		fulldirname, strerror(errno));
 	return 0;
@@ -336,36 +345,38 @@ cleanupDirectory(const char * fulldirname, const char *reldirname,
 	return 0;
     }
 
-    do {
+    for (;;) {
 	struct exclusion *e;
 
 	errno = 0;
 	ent = readdir(dir);
-	if (errno) {
+	if (errno != 0) {
 	    message(LOG_ERROR, "error reading directory entry: %s\n", 
 		    strerror(errno));
+	    (void)closedir(dir);
 	    return 0;
 	}
-	if (!ent) break;
+	if (ent == NULL)
+	    break;
 
-	if (lstat(ent->d_name, &sb)) {
-	    if (errno != ENOENT)
+	if (lstat(ent->d_name, &sb) != 0) {
+	    /* FUSE mounts by different users return EACCES by default. */
+	    if (errno != ENOENT && errno != EACCES)
 		message(LOG_ERROR, "failed to lstat %s/%s: %s\n",
 			fulldirname, ent->d_name, strerror(errno));
 	    continue;
 	}
 	
 	/* don't go crazy with the current directory or its parent */
-	if ((strcmp(ent->d_name, ".") == 0) ||
-	    (strcmp(ent->d_name, "..") == 0))
+	if (strcmp(ent->d_name, ".") == 0 || strcmp(ent->d_name, "..") == 0)
 	    continue;
       
 	/*
 	 * skip over directories named lost+found that are owned by
 	 * LOSTFOUND_UID (root)
 	 */
-	if ((strcmp(ent->d_name, "lost+found") == 0) &&
-	    S_ISDIR(sb.st_mode) && (sb.st_uid == LOSTFOUND_UID))
+	if (strcmp(ent->d_name, "lost+found") == 0 && S_ISDIR(sb.st_mode)
+	    && sb.st_uid == LOSTFOUND_UID)
 	    continue;
 
 	message(LOG_REALDEBUG, "found directory entry %s\n", ent->d_name);
@@ -389,7 +400,7 @@ cleanupDirectory(const char * fulldirname, const char *reldirname,
 		continue;
 	    p = stpcpy(full, fulldirname);
 	    p = stpcpy(p, "/");
-	    p = stpcpy(p, ent->d_name);
+	    stpcpy(p, ent->d_name);
 	    for (ep = excluded_patterns; ep != NULL; ep = ep->next) {
 		if (fnmatch(ep->pattern, full,
 			    FNM_PATHNAME | FNM_PERIOD) == 0) {
@@ -406,14 +417,14 @@ cleanupDirectory(const char * fulldirname, const char *reldirname,
 	significant_time = 0;
 	/* Set significant_time to point at the significant field of sb -
 	 * either st_atime or st_mtime depending on the flag selected. - alh */
-	if ((flags & FLAG_DIRMTIME) && S_ISDIR(sb.st_mode))
+	if ((config_flags & FLAG_DIRMTIME) != 0 && S_ISDIR(sb.st_mode))
 	    significant_time = max(significant_time, &sb.st_mtime);
 	/* The else here (and not elsewhere) is intentional */
-	else if (flags & FLAG_ATIME)
+	else if ((config_flags & FLAG_ATIME) != 0)
 	    significant_time = max(significant_time, &sb.st_atime);
-	if (flags & FLAG_MTIME)
+	if ((config_flags & FLAG_MTIME) != 0)
 	    significant_time = max(significant_time, &sb.st_mtime);
-	if (flags & FLAG_CTIME) {
+	if ((config_flags & FLAG_CTIME) != 0) {
 	    /* Even when we were told to use ctime, for directories we use
 	       mtime, because when a file in a directory is deleted, its
 	       ctime will change, and there's no way we can change it
@@ -434,19 +445,20 @@ cleanupDirectory(const char * fulldirname, const char *reldirname,
 	message(LOG_REALDEBUG, "taking as significant time: %s",
 		ctime(significant_time));
 
-	if (!sb.st_uid && !(flags & FLAG_FORCE) && !(sb.st_mode & S_IWUSR)) {
+	if (sb.st_uid == 0 && (config_flags & FLAG_FORCE) == 0
+	    && (sb.st_mode & S_IWUSR) == 0) {
 	    message(LOG_DEBUG, "non-writeable file owned by root "
 		    "skipped: %s\n", ent->d_name);;
 	    continue;
-	} else
-	    /* One more check for a different device.  Try hard not to go
-	       onto a different device.
-	    */
-	    if (sb.st_dev != st_dev || here.st_dev != st_dev) {
+	}
+	/* One more check for a different device.  Try hard not to go onto a
+	   different device. */
+	if (sb.st_dev != st_dev || here.st_dev != st_dev) {
 	    message(LOG_VERBOSE, "file on different device skipped: %s\n",
 		    ent->d_name);
 	    continue;
-	} else if (S_ISDIR(sb.st_mode)) {
+	}
+	if (S_ISDIR(sb.st_mode)) {
 	    int dd;
 
 	    if ((dd = open(".", O_RDONLY)) != -1) {
@@ -459,11 +471,11 @@ cleanupDirectory(const char * fulldirname, const char *reldirname,
 		    strcpy(full_subdir, fulldirname);
 		    strcat(full_subdir, "/");
 		    strcat(full_subdir, ent->d_name);
-		    if (cleanupDirectory(full_subdir, ent->d_name, killTime,
-					 flags, st_dev, sb.st_ino) == 0) {
+		    if (!is_bind_mount(full_subdir)
+			&& cleanupDirectory(full_subdir, ent->d_name, st_dev,
+					    sb.st_ino) == 0)
 			message(LOG_ERROR, "cleanup failed in %s: %s\n",
 				full_subdir, strerror(errno));
-		    }
 		    free(full_subdir);
 		}
 		if (fchdir(dd) != 0) {
@@ -478,10 +490,10 @@ cleanupDirectory(const char * fulldirname, const char *reldirname,
 			fulldirname, ent->d_name, strerror(errno));
 	    }
 
-	    if (*significant_time >= killTime)
+	    if (*significant_time >= kill_time)
 		continue;
 
-	    if ((flags & FLAG_FUSER) && check_fuser(ent->d_name)) {
+	    if ((config_flags & FLAG_FUSER) != 0 && check_fuser(ent->d_name)) {
 		message(LOG_VERBOSE, "file is already in use or open: %s\n",
 			ent->d_name);
 		continue;
@@ -490,13 +502,15 @@ cleanupDirectory(const char * fulldirname, const char *reldirname,
 	    /* we should try to remove the directory after cleaning up its
 	       contents, as it should contain no files.  Skip if we have
 	       specified the "no directories" flag. */
-	    if (!(flags & FLAG_NODIRS)) {
+	    if ((config_flags & FLAG_NODIRS) == 0) {
 		message(LOG_VERBOSE, "removing directory %s/%s if empty\n",
 			fulldirname, ent->d_name);
 
-		if (!(flags & FLAG_TEST)) {
+		if ((config_flags & FLAG_TEST) == 0) {
 		    if (rmdir(ent->d_name)) {
-			if (errno != ENOENT && errno != ENOTEMPTY) {
+			/* EBUSY is returned for a mount point. */
+			if (errno != ENOENT && errno != ENOTEMPTY
+			    && errno != EBUSY) {
 			    message(LOG_ERROR, "failed to rmdir %s/%s: %s\n", 
 				    fulldirname, ent->d_name, strerror(errno));
 			}
@@ -504,8 +518,14 @@ cleanupDirectory(const char * fulldirname, const char *reldirname,
 		}
 	    }
 	} else {
-	    if (*significant_time >= killTime)
-		continue;
+	    if (S_ISSOCK(sb.st_mode)) {
+		if (socket_kill_time == 0
+		    || *significant_time >= socket_kill_time)
+		    continue;
+	    } else { /* Not a socket */
+		if (*significant_time >= kill_time)
+		    continue;
+	    }
 
 #ifdef __linux
 	    /* check if it is an ext3 journal file */
@@ -536,12 +556,14 @@ cleanupDirectory(const char * fulldirname, const char *reldirname,
 	    }
 #endif
 
-	    if ((flags & FLAG_ALLFILES)
-		|| S_ISREG(sb.st_mode)
-		|| (!(flags & FLAG_NOSYMLINKS) && S_ISLNK(sb.st_mode))) {
+	    if ((config_flags & FLAG_ALLFILES) != 0
+		|| S_ISREG(sb.st_mode) || S_ISSOCK(sb.st_mode)
+		|| ((config_flags & FLAG_NOSYMLINKS) == 0
+		    && S_ISLNK(sb.st_mode))) {
 		const struct excluded_uid *u;
 
-		if ((flags & FLAG_FUSER) && check_fuser(ent->d_name)) {
+		if ((config_flags & FLAG_FUSER) != 0
+		    && check_fuser(ent->d_name)) {
 		    message(LOG_VERBOSE, "file is already in use or open: %s/%s\n",
 			    fulldirname, ent->d_name);
 		    continue;
@@ -560,14 +582,14 @@ cleanupDirectory(const char * fulldirname, const char *reldirname,
 		message(LOG_VERBOSE, "removing file %s/%s\n",
 			fulldirname, ent->d_name);
 	    
-		if (!(flags & FLAG_TEST)) {
-		    if (unlink(ent->d_name) && errno != ENOENT) 
+		if ((config_flags & FLAG_TEST) == 0) {
+		    if (unlink(ent->d_name) != 0 && errno != ENOENT) 
 			message(LOG_ERROR, "failed to unlink %s: %s\n", 
 				fulldirname, ent->d_name);
 		}
 	    }
 	}
-    } while (ent);
+    }
 
     if (closedir(dir) == -1) {
 	message(LOG_ERROR, "closedir of %s failed: %s\n",
@@ -600,10 +622,10 @@ printCopyright(void)
 static void attribute__((noreturn))
 usage(void)
 {
-    static const char msg[] = "tmpwatch [-u|-m|-c] [-MUadfqtvx] [--verbose] "
+    static const char msg[] = "tmpwatch [-u|-m|-c] [-MUXadfqtvx] [--verbose] "
 	"[--force] [--all] [--nodirs] [--nosymlinks] [--test] [--quiet] "
 	"[--atime|--mtime|--ctime] [--dirmtime] [--exclude <path>] "
-	"[--exclude-user <user>] "
+	"[--exclude-user <user>] [--exclude-pattern <pattern>] "
 #ifdef FUSER
 	"[--fuser] "
 #endif
@@ -615,6 +637,53 @@ usage(void)
     exit(1);
 }
 
+/* Set up kill_time and socket_kill_time for GRACE_MINUTES.
+
+   Connecting to an AF_UNIX socket does not update any of its times, so we
+   can't blindly remove a socket with old times - but any process listening on
+   that socket can not survive a reboot, so sockets that predate the time of
+   our boot are fair game.  We still add the grace period, both as a layer of
+   extra protection, and to let users examine contents of the temporary
+   directory for at least for some time.
+ */
+static void
+compute_kill_times(int grace_minutes)
+{
+    int grace_seconds;
+
+    grace_seconds = grace_minutes * 60;
+    message(LOG_DEBUG, "grace period is %d seconds\n", grace_seconds);
+
+    kill_time = time(NULL) - grace_seconds;
+    if ((config_flags & FLAG_ALLFILES) != 0)
+	socket_kill_time = kill_time;
+    else {
+/* We want __linux because the behavior of CLOCK_BOOTTIME is not standardized.
+   On Linux, it is the time since system boot, including the time spent in
+   sleep mode or hibernation. */
+#if defined (HAVE_CLOCK_GETTIME) && defined (CLOCK_BOOTTIME) && defined (__linux)
+	struct timespec real_clock, boot_clock;
+	time_t boot_time;
+
+	if (clock_gettime(CLOCK_REALTIME, &real_clock) != 0
+	    || clock_gettime(CLOCK_BOOTTIME, &boot_clock) != 0)
+	    message(LOG_FATAL, "Error determining boot time: %s\n",
+		    strerror(errno));
+	boot_time = real_clock.tv_sec - boot_clock.tv_sec;
+	if (real_clock.tv_nsec < boot_clock.tv_nsec)
+	    boot_time--;
+	/* We don't get the values of the two clocks at exactly the same moment,
+	   let's add a few seconds to be extra sure. */
+	boot_time -= 2;
+
+	socket_kill_time = boot_time - grace_seconds;
+#else
+	socket_kill_time = 0; /* Never remove sockets */
+#endif
+    }
+}
+
+
 int main(int argc, char ** argv)
 {
     static const struct option options[] = {
@@ -641,13 +710,14 @@ int main(int argc, char ** argv)
 
     int grace;
     char units, garbage;
-    time_t killTime;
-    int flags = 0, orig_dir;
+    int orig_dir;
     struct stat sb;
 
     set_program_name(argv[0]);
     if (argc == 1) usage();
 
+    bind_mount_init();
+
     while (1) {
 	int arg, long_index;
 
@@ -658,25 +728,25 @@ int main(int argc, char ** argv)
 
 	switch (arg) {
 	case 'M':
-	    flags |= FLAG_DIRMTIME;
+	    config_flags |= FLAG_DIRMTIME;
 	    break;
 	case 'a':
-	    flags |= FLAG_ALLFILES;
+	    config_flags |= FLAG_ALLFILES;
 	    break;
 	case 'd':
-	    flags |= FLAG_NODIRS;
+	    config_flags |= FLAG_NODIRS;
 	    break;
 	case 'f':
-	    flags |= FLAG_FORCE;
+	    config_flags |= FLAG_FORCE;
 	    break;
 	case 'l':
-	    flags |= FLAG_NOSYMLINKS;
+	    config_flags |= FLAG_NOSYMLINKS;
 	    break;
 	case 's':
-	    flags |= FLAG_FUSER;
+	    config_flags |= FLAG_FUSER;
 	    break;
 	case 't':
-	    flags |= FLAG_TEST;
+	    config_flags |= FLAG_TEST;
 	    /* fallthrough */
 	case 'v':
 	    logLevel > 0 ? logLevel -= 1 : 0;
@@ -685,13 +755,13 @@ int main(int argc, char ** argv)
 	    logLevel = LOG_FATAL;
 	    break;
 	case 'u':
-	    flags |= FLAG_ATIME;
+	    config_flags |= FLAG_ATIME;
 	    break;
 	case 'm':
-	    flags |= FLAG_MTIME;
+	    config_flags |= FLAG_MTIME;
 	    break;
 	case 'c':
-	    flags |= FLAG_CTIME;
+	    config_flags |= FLAG_CTIME;
 	    break;
 	case 'U': {
 	    struct excluded_uid *u;
@@ -758,8 +828,8 @@ int main(int argc, char ** argv)
     }
   
     /* Default to atime if neither was specified. - alh */
-    if (!(flags & (FLAG_ATIME | FLAG_MTIME | FLAG_CTIME)))
-	flags |= FLAG_ATIME;
+    if ((config_flags & (FLAG_ATIME | FLAG_MTIME | FLAG_CTIME)) == 0)
+	config_flags |= FLAG_ATIME;
 
     if (optind == argc) {
 	message(LOG_FATAL, "time (in hours) must be given\n");
@@ -790,17 +860,13 @@ int main(int argc, char ** argv)
     if (grace < 0)
 	message(LOG_FATAL, "bad time argument %s\n", argv[optind]);
 
+    compute_kill_times(grace);
+
     optind++;
     if (optind == argc) {
 	message(LOG_FATAL, "directory name(s) expected\n");
     }
 
-    grace = grace * 60;		/* minutes to seconds */
-
-    message(LOG_DEBUG, "grace period is %d seconds\n", grace);
-
-    killTime = time(NULL) - grace;
-
     /* set stdout line buffered so it is flushed before each fork */
     setvbuf(stdout, NULL, _IOLBF, 0);
 
@@ -811,7 +877,7 @@ int main(int argc, char ** argv)
 	char *path;
 
 	path = absolute_path(argv[optind], 0);
-	if (lstat(path, &sb)) {
+	if (lstat(path, &sb) != 0) {
 	    message(LOG_ERROR, "lstat() of directory %s failed: %s\n", path,
 		    strerror(errno));
 	    exit(1);
@@ -821,11 +887,9 @@ int main(int argc, char ** argv)
 	    message(LOG_DEBUG, "initial directory %s is a symlink -- "
 		    "skipping\n", path);
 	} else {
-	    if (cleanupDirectory(path, path, killTime, flags, sb.st_dev,
-				 sb.st_ino) == 0) {
+	    if (cleanupDirectory(path, path, sb.st_dev, sb.st_ino) == 0)
 		message(LOG_ERROR, "cleanup failed in %s: %s\n", path,
 			strerror(errno));
-	    }
 	    if (fchdir(orig_dir) != 0) {
 		message(LOG_FATAL, "can not return to original working "
 			"directory: %s\n", strerror(errno));
